/*
 *  TOPPERS PROJECT Home Network Working Group Software
 *
 *  Copyright (C) 2020 Cores Co., Ltd. Japan
 *
 *  上記著作権者は，以下の(1)～(4)の条件を満たす場合に限り，本ソフトウェ
 *  ア（本ソフトウェアを改変したものを含む．以下同じ）を使用・複製・改
 *  変・再配布（以下，利用と呼ぶ）することを無償で許諾する．
 *  (1) 本ソフトウェアをソースコードの形で利用する場合には，上記の著作
 *      権表示，この利用条件および下記の無保証規定が，そのままの形でソー
 *      スコード中に含まれていること．
 *  (2) 本ソフトウェアを，ライブラリ形式など，他のソフトウェア開発に使
 *      用できる形で再配布する場合には，再配布に伴うドキュメント（利用
 *      者マニュアルなど）に，上記の著作権表示，この利用条件および下記
 *      の無保証規定を掲載すること．
 *  (3) 本ソフトウェアを，機器に組み込むなど，他のソフトウェア開発に使
 *      用できない形で再配布する場合には，次のいずれかの条件を満たすこ
 *      と．
 *    (a) 再配布に伴うドキュメント（利用者マニュアルなど）に，上記の著
 *        作権表示，この利用条件および下記の無保証規定を掲載すること．
 *    (b) 再配布の形態を，別に定める方法によって，TOPPERSプロジェクトに
 *        報告すること．
 *  (4) 本ソフトウェアの利用により直接的または間接的に生じるいかなる損
 *      害からも，上記著作権者およびTOPPERSプロジェクトを免責すること．
 *      また，本ソフトウェアのユーザまたはエンドユーザからのいかなる理
 *      由に基づく請求からも，上記著作権者およびTOPPERSプロジェクトを
 *      免責すること．
 *
 *  本ソフトウェアは，無保証で提供されているものである．上記著作権者お
 *  よびTOPPERSプロジェクトは，本ソフトウェアに関して，特定の使用目的
 *  に対する適合性も含めて，いかなる保証も行わない．また，本ソフトウェ
 *  アの利用により直接的または間接的に生じたいかなる損害に関しても，そ
 *  の責任を負わない．
 *
 *  @(#) $Id: elf.c 443 2020-07-14 10:49:28Z coas-nagasima $
 */

#include <string.h>
#include <elf.h>

#define FVECT_SECT          __attribute__ ((section (".proginf")))

void _start_c(long *args);
extern int __start_text, __end_text;
extern int __start_rodata, __end_rodata;
extern int __start_data, __end_data, __start_idata;
extern int __start_bss, __end_bss;

typedef struct TProgInf {
	Elf32_Ehdr ehdr;
	Elf32_Phdr phdr[3];
	Elf32_Shdr shdr[4];
	Elf32_Shdr str_shdr;
	char shstrtab[sizeof("\0.text\0.rodata\0.data\0.bss\0.shstrtab")];
} TProgInf;

const TProgInf ProgInf FVECT_SECT = {
	{
		.e_ident = { ELFMAG0, ELFMAG1, ELFMAG2, ELFMAG3, ELFCLASS32, ELFDATA2LSB, EV_CURRENT },
		.e_type = ET_EXEC,
		.e_machine = EM_ARM,
		.e_version = EV_CURRENT,
		.e_entry = (Elf32_Addr)&_start_c,
		.e_phoff = (Elf32_Off)&((TProgInf *)0)->phdr,
		.e_shoff = (Elf32_Off)&((TProgInf *)0)->shdr,
		.e_flags = (Elf32_Word)(EF_ARM_EABI_VER5 | NT_ARM_VFP),
		.e_ehsize = (Elf32_Half)sizeof(Elf32_Ehdr),
		.e_phentsize = (Elf32_Half)sizeof(Elf32_Phdr),
		.e_phnum = (Elf32_Half)(sizeof(ProgInf.phdr) / sizeof(Elf32_Phdr)),
		.e_shentsize = (Elf32_Half)sizeof(Elf32_Shdr),
		.e_shnum = (Elf32_Half)(sizeof(ProgInf.shdr) / sizeof(Elf32_Shdr)) + 1,
		.e_shstrndx = (Elf32_Half)(sizeof(ProgInf.shdr) / sizeof(Elf32_Shdr)),
	},
	{
		{
			.p_type = (Elf32_Word)PT_LOAD,
			.p_offset = (Elf32_Off)&((TProgInf *)0)->phdr[0],
			.p_vaddr = (Elf32_Addr)&__start_text,
			.p_paddr = (Elf32_Addr)&__start_text,
			.p_filesz = (Elf32_Word)&__end_text/* - (Elf32_Addr)&__start_text*/,
			.p_memsz = (Elf32_Word)&__end_text/* - (Elf32_Addr)&__start_text*/,
			.p_flags = (Elf32_Word)(PF_X | PF_R),
			.p_align = (Elf32_Word)8,
		},
		{
			.p_type = (Elf32_Word)PT_LOAD,
			.p_offset = (Elf32_Off)&((TProgInf *)0)->phdr[1],
			.p_vaddr = (Elf32_Addr)&__start_rodata,
			.p_paddr = (Elf32_Addr)&__start_rodata,
			.p_filesz = (Elf32_Word)&__end_rodata/* - (Elf32_Addr)&__start_rodata*/,
			.p_memsz = (Elf32_Word)&__end_rodata/* - (Elf32_Addr)&__start_rodata*/,
			.p_flags = (Elf32_Word)(PF_R),
			.p_align = (Elf32_Word)8,
		},
		{
			.p_type = (Elf32_Word)PT_LOAD,
			.p_offset = (Elf32_Off)&((TProgInf *)0)->phdr[2],
			.p_vaddr = (Elf32_Addr)&__start_data,
			.p_paddr = (Elf32_Addr)&__start_idata,
			.p_filesz = (Elf32_Word)&__end_data/* - (Elf32_Addr)&__start_data*/,
			.p_memsz = (Elf32_Word)&__end_data/* - (Elf32_Addr)&__start_data*/,
			.p_flags = (Elf32_Word)(PF_W | PF_R),
			.p_align = (Elf32_Word)8,
		}
	},
	{
		{
			.sh_name = (Elf32_Word)1,
			.sh_type = (Elf32_Word)SHT_PROGBITS,
			.sh_flags = (Elf32_Word)(SHF_ALLOC | SHF_EXECINSTR),
			.sh_addr = (Elf32_Addr)&__start_text,
			.sh_offset = (Elf32_Off)&__start_text/* - (Elf32_Off)&ProgInf*/,
			.sh_size = (Elf32_Word)&__end_text/* - (Elf32_Off)&__start_text*/,
			.sh_link = (Elf32_Word)0,
			.sh_info = (Elf32_Word)0,
			.sh_addralign = (Elf32_Word)8,
			.sh_entsize = (Elf32_Word)0,
		},
		{
			.sh_name = (Elf32_Word)7,
			.sh_type = (Elf32_Word)SHT_PROGBITS,
			.sh_flags = (Elf32_Word)(SHF_ALLOC),
			.sh_addr = (Elf32_Addr)&__start_rodata,
			.sh_offset = (Elf32_Off)&__start_rodata/* - (Elf32_Off)&ProgInf*/,
			.sh_size = (Elf32_Word)&__end_rodata/* - (Elf32_Off)&__start_rodata*/,
			.sh_link = (Elf32_Word)0,
			.sh_info = (Elf32_Word)0,
			.sh_addralign = (Elf32_Word)8,
			.sh_entsize = (Elf32_Word)0,
		},
		{
			.sh_name = (Elf32_Word)15,
			.sh_type = (Elf32_Word)SHT_PROGBITS,
			.sh_flags = (Elf32_Word)(SHF_ALLOC | SHF_WRITE),
			.sh_addr = (Elf32_Addr)&__start_data,
			.sh_offset = (Elf32_Off)&__start_idata/* - (Elf32_Off)&ProgInf*/,
			.sh_size = (Elf32_Word)&__end_data/* - (Elf32_Off)&__start_data*/,
			.sh_link = (Elf32_Word)0,
			.sh_info = (Elf32_Word)0,
			.sh_addralign = (Elf32_Word)8,
			.sh_entsize = (Elf32_Word)0,
		},
		{
			.sh_name = (Elf32_Word)21,
			.sh_type = (Elf32_Word)SHT_NOBITS,
			.sh_flags = (Elf32_Word)(SHF_ALLOC | SHF_WRITE),
			.sh_addr = (Elf32_Addr)&__start_bss,
			.sh_offset = (Elf32_Off)&__start_bss/* - (Elf32_Off)&ProgInf*/,
			.sh_size = (Elf32_Word)&__end_bss/* - (Elf32_Off)&__start_bss*/,
			.sh_link = (Elf32_Word)0,
			.sh_info = (Elf32_Word)0,
			.sh_addralign = (Elf32_Word)8,
			.sh_entsize = (Elf32_Word)0,
		}
	},
	{
		.sh_name = (Elf32_Word)26,
		.sh_type = (Elf32_Word)SHT_STRTAB,
		.sh_flags = (Elf32_Word)0,
		.sh_addr = (Elf32_Addr)0,
		.sh_offset = (Elf32_Off)&((TProgInf *)0)->shstrtab,
		.sh_size = (Elf32_Word)sizeof(ProgInf.shstrtab),
		.sh_link = (Elf32_Word)0,
		.sh_info = (Elf32_Word)0,
		.sh_addralign = (Elf32_Word)1,
		.sh_entsize = (Elf32_Word)0,
	},
	"\0.text\0.rodata\0.data\0.bss\0.shstrtab"
};
